# -*- coding: utf-8 -*-

# FOGLAMP_BEGIN
# See: https://foglamp-foglamp-documentation.readthedocs-hosted.com
# FOGLAMP_END

""" Server functions for plugin module to detect bad bearing """


__author__ = "Douglas Orr"
__copyright__ = "Copyright (c) 2020 Dianomic Systems"
__license__ = "Apache 2.0"
__version__ = "${VERSION}"

import os
import sys
import threading
import traceback
import copy
import time
import hashlib
import glob
from datetime import datetime
from contextlib import suppress
import logging

import cv2

from foglamp.common import logger
from foglamp.common.plugin_helpers import PluginRPCServer

_LOGGER = logger.setup(__name__, level=logging.INFO)


class Webcam(threading.Thread):

    def __init__(self, parent, camera_number, fps):
        """[pull images from a webcam every <period>, rotate out old, unprocessed images]

        Args:
            parent ([type]): [The parent class that contols the this thread]
            fps ([type]): [the frames per minute]
            camera_number (int, optional): [The camera number that corresponds to 
            /dev/videoi where i is the camera number.]. Defaults to 0.
        """
        super().__init__()

        self.parent = parent  # for locks and synchronizatoin
        self.logger = parent.logger

        self.camera_number = None
        self.fps = None
        self.vid = None
        
        self._setup_camera(camera_number, fps)

    def _setup_camera(self, camera_number, fps):
        """ _setup_camera - create a valid self.vid video capture """
        # CALLED WITH self.update held
        self.logger.info("_setup camera: {}->{}  {}->{}".format(
            self.camera_number, camera_number,
            self.fps, fps
        ))
        if self.camera_number != camera_number:
            if self.vid is not None:
                self.vid.release()  # need to get a new cv object
            self.camera_number = camera_number
            self.vid = cv2.VideoCapture(self.camera_number)

            if not self.vid.isOpened():
                self.camera_number = None
                self.vid = None
                raise ValueError(
                    "Could not open camera {}".format(camera_number)
                )

        self.fps = fps

    def _disable_camera(self):
        # CALLED WITH self.update held
        if self.vid is not None:
            self.logger.info("RELEASE camera")
            self.vid.release()
            self.vid = None
        
    def run(self):
        """[The main thread that keeps on fetching frames and goes to
         sleep for a configurable amount of time.]

        Raises:
            ValueError: [if opencv could get open the camera with camera id provided.]
            
        """
        
        while True:

            with self.parent.update:
                # check to see if we've had a config update
                if self.parent.updated:
                    self._setup_camera(self.parent.camera_number, self.parent.fps)
                    self.parent.updated = False

                if self.parent.exit_webcam:
                    self._disable_camera()
                    break

            # read webcam images and queue them for processing
            _start = datetime.utcnow()

            if self.vid is not None:
                # Capture the video frame by frame
                ret, frame = self.vid.read()

                if ret:
                    self.queue_frame(frame)

            # sleep the rest of the time slot
            _elapsed = (datetime.utcnow() - _start).total_seconds()
            _rem = (1 / self.fps) - _elapsed
            self.logger.info("sleeping fps:{} for {}".format(self.fps, _rem))
            if _rem > 0:
                time.sleep(_rem)

    def queue_frame(self, frame):
        """[Queue a frame that will be the next frame to be written to a jpg file.]

        Args:
            frame ([type]): [a frame to be queued]
        """
        p = self.parent

        _t = datetime.utcnow()
        with p.queue:
            p.frames.append((_t, frame))

            if len(p.frames) > p.max_image_queue:
                self.logger.debug("PRODUCE: overflow")
                p.frames.pop(0)

            p.queue.notify()  # signal to processing thread


class WebcamServer(PluginRPCServer):
    """[create a queue of images;
    extract images file names when polled and send them upstream as an asset]
    """
    def __init__(self):
        super().__init__("server")

        self.logger = _LOGGER
        
        # slots 
        self.camera_number = None
        self.fpm = None
        self.full_media_dir = None

        self.max_image_queue = 10

        self.queue = (
            threading.Condition()
        )  # lock/condition for incoming work/results
        self.frames = []

        self.update = threading.Condition()  # lock/condition for changes in config
        self.updated = False
        self.webcam = None
        self.exit_webcam = False

    def _update_camera(self):
        """ _start_camera_capture - turn on camera and start recording images in new thread """
        if self.webcam is None:
            # start a webcam thread
            self.webcam = Webcam(self, self.camera_number, self.fps)
            self.webcam.daemon = True  # exit if program exits
            self.webcam.start()
        else:
            # update existing webcam thread
            with self.update:
                self.updated = True

    # ### PLUGIN API ### #

    def plugin_init(self, config):
        """[Plugin init at the server side.]

        Args:
            config ([dict]): [A configuration from client side.]

        Returns:
            [dict]: [returns the same config as it is.]
        """
        # called to initialize or reinitialize camera
        self.logger.info("PLUGIN INIT")
        
        self.config_update(config)   # transfer config elements to handle

        if self.full_media_dir != config["full_media_dir"]:
            self.full_media_dir = config["full_media_dir"]

            # # # prepare the folder # # #
            os.makedirs(self.full_media_dir, exist_ok=True)

        self.fps = self.fpm / 60

        self._update_camera()

    def plugin_poll(self):
        """[pulls and stores a list of images, from webcam.]

        Returns:
            [list]: [A list of image  file names]
        """
        if self.webcam is None:
            return None

        # extract frames that have been captured by the background thread
        _frames = []
        with self.queue:
            while len(_frames) == 0:
                if len(self.frames) == 0:
                    self.queue.wait()
                else:
                    _frames = self.frames
                    self.frames = []

        # convert frames into files, return a "reading" that has the base file names
        # (there should be a final filter that will delete them if they haven't been preserved
        #  by some intermediate filter)
        _results = []
        for t, frame in _frames:
            _fname = "CAM-{}.jpg".format(t.strftime("%Y.%m.%d.%H.%M.%S.%f"))
            try:
                cv2.imwrite(os.path.join(self.full_media_dir, _fname), frame)
            except:
                traceback.print_exc()
            else:
                _results.append(_fname)

        return _results

    def plugin_shutdown(self):
        """[shutdown at the server side.]
        """
        self.logger.warning("Server shutting down")

        with self.update:
            if self.webcam is not None:
                self.exit_webcam = True  # signal any stubborn threads to quit
                self.webcam.updated = True

        if self.exit_webcam:
            self.webcam.join()

        self.webcam = None
        self.exit_webcam = False


_LOGGER.info("Web Cam media server SERVING....")
rpc = WebcamServer()
rpc.serve()
